import datetime
import random

import pytest

from syrupy.extensions.json import JSONSnapshotExtension
from syrupy.matchers import path_type


@pytest.fixture
def snapshot_json(snapshot):
    return snapshot.use_extension(JSONSnapshotExtension)


def test_matcher(snapshot_json):
    content = {
        "int": random.randint(1, 100),
        "date": datetime.datetime.utcnow(),
        "foo": {
            "x": "y",
            "another_date": datetime.datetime.utcnow(),
        },
    }
    matcher = path_type(
        {
            "int": (int,),
            "date": (datetime.date,),
            "foo.another_date": (dict, datetime.datetime),
        }
    )
    assert snapshot_json(matcher=matcher) == content



from pathlib import Path

import pytest


@pytest.fixture
def testcases():
    return {
        "base": (
            """
            def test_a(snapshot):
                assert snapshot(name="xyz") == "case 1"
                assert snapshot(name="zyx") == "case 2"
            """
        ),
        "modified": (
            """
            def test_a(snapshot):
                assert snapshot(name="xyz") == "case 1"
                assert snapshot(name="zyx") == "case ??"
            """
        ),
    }


@pytest.fixture
def run_testcases(pytester, testcases) -> tuple[pytest.Pytester, dict[str, str]]:
    pytester.makepyfile(test_1=testcases["base"])
    result = pytester.runpytest(
        "-v",
        "--snapshot-dirname=snaps",
        "--snapshot-update",
    )
    result.stdout.re_match_lines((r"2 snapshots generated\.",))
    return pytester, testcases


def test_run_all(run_testcases, plugin_args_fails_xdist):
    pytester, _ = run_testcases
    result = pytester.runpytest(
        "-v", "--snapshot-dirname=snaps", *plugin_args_fails_xdist
    )
    result.stdout.re_match_lines(("2 snapshots passed",))
    assert result.ret == 0

    # Assert there's a snaps folder but no __snapshots__ folder
    expected_dir = pytester.path / Path("snaps")
    unexpected_dir = pytester.path / Path("__snapshots__")

    assert expected_dir.exists()
    assert not unexpected_dir.exists()


def test_failure(run_testcases, plugin_args_fails_xdist):
    pytester, testcases = run_testcases
    pytester.makepyfile(test_1=testcases["modified"])
    result = pytester.runpytest(
        "-vv", "--snapshot-dirname=snaps", *plugin_args_fails_xdist
    )
    result.stdout.re_match_lines(("1 snapshot failed. 1 snapshot passed.",))
    assert result.ret == 1

    # Assert there's a snaps folder but no __snapshots__ folder
    expected_dir = pytester.path / Path("snaps")
    unexpected_dir = pytester.path / Path("__snapshots__")

    assert expected_dir.exists()
    assert not unexpected_dir.exists()


def test_update(run_testcases, plugin_args_fails_xdist):
    pytester, testcases = run_testcases
    pytester.makepyfile(test_1=testcases["modified"])
    result = pytester.runpytest(
        "-v", "--snapshot-dirname=snaps", "--snapshot-update", *plugin_args_fails_xdist
    )
    assert "Can not relate snapshot name" not in str(result.stdout)
    result.stdout.re_match_lines(("1 snapshot passed. 1 snapshot updated.",))
    assert result.ret == 0

    # Assert there's a snaps folder but no __snapshots__ folder
    expected_dir = pytester.path / Path("snaps")
    unexpected_dir = pytester.path / Path("__snapshots__")

    assert expected_dir.exists()
    assert not unexpected_dir.exists()



import pytest

from syrupy.extensions.single_file import SingleFileAmberSnapshotExtension


@pytest.fixture
def snapshot_single(snapshot):
    return snapshot.use_extension(SingleFileAmberSnapshotExtension)


def test_amber_single_file(snapshot_single):
    assert snapshot_single == 1
    assert snapshot_single == {"a": "b"}
    assert (
        snapshot_single
        == """
        Multi
        line
        string
    """
    )



import pytest

_TEST = """
def test_foo(snapshot):
    assert {**base} == snapshot(name="a")
    assert {**base, **extra} == snapshot(name="b", diff="a")
"""


def _make_file(testdir, base, extra):
    testdir.makepyfile(
        test_file="\n\n".join([f"base = {base!r}", f"extra = {extra!r}", _TEST])
    )


def _run_test(testdir, base, extra, expected_update_lines):
    _make_file(testdir, base=base, extra=extra)

    # Run with --snapshot-update, to generate/update snapshots:
    result = testdir.runpytest(
        "-v",
        "--snapshot-update",
    )
    result.stdout.re_match_lines((expected_update_lines,))
    assert result.ret == 0

    # Run without --snapshot-update, to validate the snapshots are actually up-to-date
    result = testdir.runpytest("-v")
    result.stdout.re_match_lines((r"2 snapshots passed\.",))
    assert result.ret == 0


def test_diff_lifecycle(testdir) -> pytest.Testdir:
    # first: create both snapshots completely from scratch
    _run_test(
        testdir,
        base={"A": 1},
        extra={"X": 10},
        expected_update_lines=r"2 snapshots generated\.",
    )

    # second: edit the base data, to change the data for both snapshots (only changes the serialized output for the base snapshot `a`).
    _run_test(
        testdir,
        base={"A": 1, "B": 2},
        extra={"X": 10},
        expected_update_lines=r"1 snapshot passed. 1 snapshot updated\.",
    )

    # third: edit just the extra data (only changes the serialized output for the diff snapshot `b`)
    _run_test(
        testdir,
        base={"A": 1, "B": 2},
        extra={"X": 10, "Y": 20},
        expected_update_lines=r"1 snapshot passed. 1 snapshot updated\.",
    )



from typing import (
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from syrupy.types import (
        PropertyFilter,
        PropertyName,
        PropertyPath,
    )


def paths(*path_parts: str) -> "PropertyFilter":
    """
    Factory to create a filter using list of path strings.

    This filter does not work well when combined with "include" and
    nested paths, since "include" oeprates per key as an object is traversed
    for serialization. For nested paths, we must include all parents. To accomplish
    this, we provide an alternative "paths_include" filter which does this
    automatically.
    """

    if not path_parts:
        raise TypeError("At least 1 path argument is required.")

    parts: set[str] = set(path_parts)

    def path_filter(*, prop: "PropertyName", path: "PropertyPath") -> bool:
        path_str = ".".join(str(p) for p, _ in (*path, (prop, None)))
        return path_str in parts

    return path_filter


def paths_include(*path_parts: tuple[str, ...] | list[str]) -> "PropertyFilter":
    """
    Factory to create a filter using list of path tuples.
    """

    if not path_parts:
        raise TypeError("At least 1 path argument is required.")

    # "include" operates per key as an object is traversed for serialization.
    # This means, if matching a nested path, we must also include all parents.
    parts: set[tuple[str, ...]] = set()
    for path_part in path_parts:
        if isinstance(path_part, (list, tuple)):
            for idx in range(len(path_part)):
                parts.add(tuple(path_part[: idx + 1]))
        else:
            raise TypeError("Unexpected argument. Expected list/tuple.")

    def path_filter(*, prop: "PropertyName", path: "PropertyPath") -> bool:
        path_tuple = tuple(str(p) for p, _ in (*path, (prop, None)))
        return path_tuple in parts

    return path_filter


def props(*prop_names: str) -> "PropertyFilter":
    """
    Factory to create filter using list of props
    """

    if not prop_names:
        raise TypeError("At least 1 prop name is required.")

    def prop_filter(*, prop: "PropertyName", path: "PropertyPath") -> bool:
        return any(str(prop) == p for p in prop_names)

    return prop_filter



from pathlib import Path
from unittest.mock import MagicMock

import pytest

from syrupy.constants import PYTEST_NODE_SEP
from syrupy.location import PyTestLocation


def mock_pytest_item(node_id: str, method_name: str) -> "pytest.Item":
    mock_node = MagicMock(spec=pytest.Item)
    mock_node.nodeid = node_id
    [filepath, *_, nodename] = node_id.split(PYTEST_NODE_SEP)
    mock_node.name = nodename
    mock_node.path = Path(filepath)
    mock_node.obj = MagicMock()
    mock_node.obj.__module__ = Path(filepath).stem
    mock_node.obj.__name__ = method_name
    return mock_node


@pytest.mark.parametrize(
    "node_id, method_name, expected_filename, expected_classname, expected_snapshotname",
    (
        (
            "/tests/module/test_file.py::TestClass::method_name",
            "method_name",
            "test_file",
            "TestClass",
            "TestClass.method_name",
        ),
        (
            "/tests/module/test_file.py::TestClass::method_name[1]",
            "method_name",
            "test_file",
            "TestClass",
            "TestClass.method_name[1]",
        ),
        (
            "/tests/module/nest/test_file.py::TestClass::TestSubClass::method_name",
            "method_name",
            "test_file",
            "TestClass.TestSubClass",
            "TestClass.TestSubClass.method_name",
        ),
    ),
)
def test_location_properties(
    node_id,
    method_name,
    expected_filename,
    expected_classname,
    expected_snapshotname,
):
    location = PyTestLocation(mock_pytest_item(node_id, method_name))
    assert location.classname == expected_classname
    assert location.basename == expected_filename
    assert location.snapshot_name == expected_snapshotname


@pytest.mark.parametrize(
    "node_id, method_name,"
    "expected_location_matches, expected_location_misses,"
    "expected_snapshot_matches, expected_snapshot_misses",
    (
        (
            "/tests/module/test_file.py::TestClass::method_name",
            "method_name",
            ("test_file.snap", "__snapshots__/test_file", "test_file/1.snap"),
            (
                "test.snap",
                "__others__/test/file.snap",
                "test_file_extra.snap",
                "__snapshots__/test_file_extra",
                "test_file_extra/1.snap",
                "test_file/extra/1.snap",
                "__snapshots__/test_file/extra/even/more/1.snap",
            ),
            (
                "TestClass.method_name",
                "TestClass.method_name[1]",
                "TestClass.method_name.1",
            ),
            ("method_name", "TestClass.method_names"),
        ),
        (
            "/tests/module/test_file.py::TestClass::method_name[1]",
            "method_name",
            (
                "test_file.snap",
                "__snapshots__/test_file",
                "test_file/TestClass.method_name[1].snap",
                "test_file/TestClass.method_name[1].1.snap",
                "test_file/TestClass.method_name[1][1].snap",
            ),
            (
                "test.snap",
                "__others__/test/file.snap",
                "test_file_extra.snap",
                "__snapshots__/test_file_extra",
                "test_file_extra/1.snap",
                "test_file/extra/1.snap",
                "__snapshots__/test_file/extra/even/more/1.snap",
                "test_file/TestClass.method_name[1]xyz.snap",
                "test_file/TestClass.method_name[2].snap",
            ),
            (
                "TestClass.method_name",
                "TestClass.method_name[1]",
                "TestClass.method_name.1",
                "TestClass.method_name[1][1]",
                "TestClass.method_name[1].1",
            ),
            ("method_name", "TestClass.method_names"),
        ),
    ),
)
def test_location_matching(
    node_id,
    method_name,
    expected_location_matches,
    expected_location_misses,
    expected_snapshot_matches,
    expected_snapshot_misses,
):
    location = PyTestLocation(mock_pytest_item(node_id, method_name))

    for location_match in expected_location_matches:
        assert location.matches_snapshot_location(location_match)

    for location_miss in expected_location_misses:
        assert not location.matches_snapshot_location(location_miss)

    for snapshot_match in expected_snapshot_matches:
        assert location.matches_snapshot_name(snapshot_match)

    for snapshot_miss in expected_snapshot_misses:
        assert not location.matches_snapshot_name(snapshot_miss)



"""Pydantic plugin for Syrupy Amber serializer."""

from typing import Any

from pydantic import BaseModel

from syrupy.extensions.amber.serializer import (
    AmberDataSerializer,
    AmberDataSerializerPlugin,
    attr_getter,
)
from syrupy.types import SerializableData


class PydanticPlugin(AmberDataSerializerPlugin):
    """A Syrupy extension that serializes Pydantic models using Amber format."""

    @classmethod
    def is_data_serializable(cls, data: "SerializableData") -> bool:
        """Check if the data is a Pydantic BaseModel instance."""
        return isinstance(data, BaseModel)

    @classmethod
    def serialize(cls, data: BaseModel, **kwargs: Any) -> str:
        """Serialize a Pydantic BaseModel instance into Amber format.

        Uses the model's defined fields to ensure consistent ordering.
        """
        keys = [name for name, _ in data]
        return AmberDataSerializer.serialize_custom_iterable(
            data=data,
            resolve_entries=(keys, attr_getter, None),
            separator="=",
            **kwargs,
        )



import pytest


@pytest.fixture
def testcases():
    return {
        "used": (
            """
            def test_used(snapshot):
                assert snapshot == 'used'
            """
        ),
        "unused": (
            """
            def test_unused(snapshot):
                assert snapshot == 'unused'
            """
        ),
    }


@pytest.fixture
def extra_testcases():
    return {
        "extra_a": (
            """
            def test_extra_a(snapshot):
                assert snapshot == 'extra_a'
            """
        ),
        "extra_b": (
            """
            def test_extra_b(snapshot):
                assert snapshot == 'extra_b'
            """
        ),
    }


@pytest.fixture
def run_testfiles_with_update(testdir):
    def run_testfiles_with_update_impl(**testfiles):
        testdir.makepyfile(
            **{
                filename: "\n\n".join(cases.values())
                for (filename, cases) in testfiles.items()
            }
        )
        result = testdir.runpytest("-v", "--snapshot-update")
        result.stdout.re_match_lines((r"[0-9]+ snapshots generated\.",))
        return testdir

    return run_testfiles_with_update_impl


@pytest.mark.parametrize(
    (
        "options",
        "expected_status_code",
    ),
    (
        (("-v", "--snapshot-details"), 1),
        (("-v", "--snapshot-details", "--snapshot-warn-unused"), 0),
    ),
)
def test_unused_snapshots_details(
    options,
    expected_status_code,
    run_testfiles_with_update,
    testcases,
    plugin_args_fails_xdist,
):
    testdir = run_testfiles_with_update(test_file=testcases)
    testdir.makepyfile(test_file=testcases["used"])

    result = testdir.runpytest(*options, *plugin_args_fails_xdist)
    result.stdout.re_match_lines(
        (
            r"1 snapshot passed\. 1 snapshot unused\.",
            r"Unused test_unused \(__snapshots__[\\/]test_file.ambr\)",
            r"Re-run pytest with --snapshot-update to delete unused snapshots\.",
        )
    )
    assert result.ret == expected_status_code


def test_unused_snapshots_details_multiple_tests(
    run_testfiles_with_update, testcases, extra_testcases, plugin_args_fails_xdist
):
    testdir = run_testfiles_with_update(
        test_file=testcases, test_second_file=extra_testcases
    )
    testdir.makepyfile(
        test_file="\n\n".join(testcases.values()),
        test_second_file="",
    )

    result = testdir.runpytest("-v", "--snapshot-details", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(
        (
            r"2 snapshots passed\. 2 snapshots unused\.",
            r"Unused test_extra_a, test_extra_b "
            r"\(__snapshots__[\\/]test_second_file.ambr\)",
            r"Re-run pytest with --snapshot-update to delete unused snapshots\.",
        )
    )
    assert result.ret == 1


def test_unused_snapshots_details_multiple_locations(
    run_testfiles_with_update, testcases, extra_testcases, plugin_args_fails_xdist
):
    testdir = run_testfiles_with_update(
        test_file=testcases, test_second_file=extra_testcases
    )
    testdir.makepyfile(
        test_file=testcases["used"],
        test_second_file=extra_testcases["extra_a"],
    )

    result = testdir.runpytest("-v", "--snapshot-details", *plugin_args_fails_xdist)
    result.stdout.re_match_lines_random(
        (
            r"2 snapshots passed\. 2 snapshots unused\.",
            r"Unused test_extra_b \(__snapshots__[\\/]test_second_file.ambr\)",
            r"Unused test_unused \(__snapshots__[\\/]test_file.ambr\)",
            r"Re-run pytest with --snapshot-update to delete unused snapshots\.",
        )
    )
    assert result.ret == 1


def test_unused_snapshots_details_no_details_on_deletion(
    run_testfiles_with_update, testcases, plugin_args_fails_xdist
):
    testdir = run_testfiles_with_update(test_file=testcases)
    testdir.makepyfile(test_file=testcases["used"])

    result = testdir.runpytest(
        "-v", "--snapshot-details", "--snapshot-update", *plugin_args_fails_xdist
    )
    result.stdout.re_match_lines(
        (
            r"1 snapshot passed\. 1 unused snapshot deleted\.",
            r"Deleted test_unused \(__snapshots__[\\/]test_file.ambr\)",
        )
    )
    assert result.ret == 0


def test_created_and_updates_details(testdir, plugin_args_fails_xdist):
    # Generate initial snapshots.
    testdir.makepyfile(
        test_generated="""
            def test_generated_1(snapshot):
                assert snapshot == "1"

            def test_generated_2(snapshot):
                assert snapshot == "2"
            """
    )
    result = testdir.runpytest("-v", "--snapshot-update")
    result.stdout.re_match_lines((r"2 snapshots generated\.",))

    # Update one of the snapshots and generate a new one
    testdir.makepyfile(
        test_generated="""
            def test_generated_1(snapshot):
                assert snapshot == "1"

            def test_generated_2(snapshot):
                assert snapshot == "2-update"

            def test_generated_3(snapshot):
                assert snapshot == "3"
            """
    )

    result = testdir.runpytest(
        "-v", "--snapshot-details", "--snapshot-update", *plugin_args_fails_xdist
    )
    result.stdout.re_match_lines(
        (
            r"1 snapshot passed\. 1 snapshot generated\. 1 snapshot updated\.",
            r"Generated test_generated_3 \(__snapshots__[\\/]test_generated.ambr\)",
            r"Updated test_generated_2 \(__snapshots__[\\/]test_generated.ambr\)",
        )
    )
    assert result.ret == 0



from invoke import task

from .utils import ctx_run


@task(
    help={
        "coverage": "Build and report on test coverage",
        "test-pattern": "Pattern used to select test files to run",
        "update-snapshots": "Create, update or delete snapshot files",
        "verbose": "Verbose output e.g. non captured logs etc.",
    }
)
def test(
    ctx,
    coverage=False,
    test_pattern=None,
    update_snapshots=False,
    verbose=False,
    debug=False,
):
    """
    Run entire test suite
    """
    flags = {
        "-s -vv": verbose,
        f"-k {test_pattern}": test_pattern,
        "--snapshot-update": update_snapshots,
    }
    coverage_module = "coverage run -m " if coverage else ""
    test_flags = " ".join(flag for flag, enabled in flags.items() if enabled)

    if debug and coverage:
        raise Exception("The debug and coverage options are mutually exclusive.")

    if debug:
        ctx_run(
            ctx,
            f"python -m debugpy --listen 5678 --wait-for-client -m pytest {test_flags} ./tests",
        )
    else:
        ctx_run(ctx, f"python -m {coverage_module}pytest {test_flags} ./tests")

    if coverage:
        ctx_run(ctx, "coverage lcov -o coverage/coverage.lcov")



class ReprWithNewline:
    def __init__(self, newlines: int = 1) -> None:
        self.newlines = newlines

    def __repr__(self) -> str:
        newlines = "\n" * self.newlines
        return f"ReprWithNewline{newlines}"


def test_trailing_no_newline_in_repr(snapshot):
    assert ReprWithNewline(0) == snapshot


def test_trailing_newline_in_repr(snapshot):
    assert ReprWithNewline(1) == snapshot


def test_trailing_2_newlines_in_repr(snapshot):
    assert ReprWithNewline(2) == snapshot


class MultilineRepr:
    def __repr__(self) -> str:
        return "\n".join(
            [
                "Line1",
                "Line2\n",  # extra newline
                "Line3 ",  # with an extra space
            ]
        )


def test_multiline_repr(snapshot):
    assert MultilineRepr() == snapshot



def test_raises_informative_type_error_when_serializing_non_bytes(testdir):
    testdir.makepyfile(
        test_file="""
        from syrupy.extensions.single_file import SingleFileSnapshotExtension

        def test_case(snapshot):
            assert {"x": 0} == snapshot(extension_class=SingleFileSnapshotExtension)
        """
    )

    result = testdir.runpytest("-v")
    assert result.ret == 1
    result.stdout.re_match_lines(
        (r".*TypeError: Can't serialize.*You must convert the data.*",)
    )



from pathlib import Path

import pytest


# EqualsAssertionError comes from:
# https://github.com/JetBrains/intellij-community/blob/cd9bfbd98a7dca730fbc469156ce1ed30364afba/python/helpers/pycharm/teamcity/diff_tools.py#L53
@pytest.fixture
def mock_teamcity_diff_tools(testdir: "pytest.Testdir"):
    teamcity_pkg = testdir.mkpydir("teamcity")
    diff_tools_file = teamcity_pkg / Path("diff_tools.py")
    diff_tools_file.write_text(
        """
class EqualsAssertionError:
    def __init__(self, expected, actual, msg=None, preformated=False, real_exception=None): # noqa: E501
        self.real_exception = real_exception
        self.expected = expected
        self.actual = actual
        self.msg = str(msg)
    """,
        "utf-8",
    )


@pytest.mark.filterwarnings("default")
def test_logs_a_warning_if_unable_to_apply_patch(testdir, plugin_args):
    testdir.makepyfile(
        test_file="""
    def test_case(snapshot):
        assert snapshot == [1, 2]
    """
    )
    testdir.runpytest("-v", "--snapshot-update", *plugin_args)
    testdir.makepyfile(
        test_file="""
    def test_case(snapshot):
        assert snapshot == [1, 2, 3]
    """
    )

    result = testdir.runpytest("-v", "--snapshot-patch-pycharm-diff", *plugin_args)
    result.assert_outcomes(failed=1, passed=0, warnings=1)


@pytest.mark.filterwarnings("default")
def test_patches_pycharm_diff_tools_when_flag_set(
    testdir, mock_teamcity_diff_tools, plugin_args
):
    # Generate initial snapshot
    testdir.makepyfile(
        test_file="""
    def test_case(snapshot):
        assert snapshot == [1, 2]
    """
    )
    testdir.runpytest("-v", "--snapshot-update", *plugin_args)

    # Generate diff and mimic EqualsAssertionError being thrown
    testdir.makepyfile(
        test_file="""

    def test_case(snapshot):
        try:
            assert snapshot == [1, 2, 3]
        except:
            from teamcity.diff_tools import EqualsAssertionError

            err = EqualsAssertionError(expected=snapshot, actual=[1,2,3])
            print("Expected:", repr(err.expected))
            print("Actual:", repr(err.actual))
            raise
    """
    )

    result = testdir.runpytest("-v", "--snapshot-patch-pycharm-diff", *plugin_args)
    # No warnings because patch should have been successful
    result.assert_outcomes(failed=1, passed=0, warnings=0)

    result.stdout.re_match_lines(
        (
            r"Expected: 'list([\n  1,\n  2,\n])'",
            # Actual is the amber-style list representation
            r"Actual: 'list([\n  1,\n  2,\n  3,\n])'",
        )
    )


@pytest.mark.filterwarnings("default")
def test_patches_pycharm_diff_tools_when_flag_set_and_snapshot_on_right(
    testdir, mock_teamcity_diff_tools, plugin_args
):
    # Generate initial snapshot
    testdir.makepyfile(
        test_file="""
    def test_case(snapshot):
        assert [1, 2] == snapshot
    """
    )
    testdir.runpytest("-v", "--snapshot-update", *plugin_args)

    # Generate diff and mimic EqualsAssertionError being thrown
    testdir.makepyfile(
        test_file="""

    def test_case(snapshot):
        try:
            assert [1, 2, 3] == snapshot
        except:
            from teamcity.diff_tools import EqualsAssertionError

            err = EqualsAssertionError(expected=[1,2,3], actual=snapshot)
            print("Expected:", repr(err.expected))
            print("Actual:", repr(err.actual))
            raise
    """
    )

    result = testdir.runpytest("-v", "--snapshot-patch-pycharm-diff", *plugin_args)
    # No warnings because patch should have been successful
    result.assert_outcomes(failed=1, passed=0, warnings=0)

    result.stdout.re_match_lines(
        (
            r"Expected: 'list([\n  1,\n  2,\n])'",
            # Actual is the amber-style list representation
            r"Actual: 'list([\n  1,\n  2,\n  3,\n])'",
        )
    )


@pytest.mark.filterwarnings("default")
def test_it_does_not_patch_pycharm_diff_tools_by_default(
    testdir, mock_teamcity_diff_tools, plugin_args
):
    # Generate initial snapshot
    testdir.makepyfile(
        test_file="""
    def test_case(snapshot):
        assert snapshot == [1, 2]
    """
    )
    testdir.runpytest("-v", "--snapshot-update", *plugin_args)

    # Generate diff and mimic EqualsAssertionError being thrown
    testdir.makepyfile(
        test_file="""

    def test_case(snapshot):
        try:
            assert snapshot == [1, 2, 3]
        except:
            from teamcity.diff_tools import EqualsAssertionError

            err = EqualsAssertionError(expected=snapshot, actual=[1,2,3])
            print("Expected:", repr(str(err.expected)))
            print("Actual:", repr(str(err.actual)))
            raise
    """
    )

    result = testdir.runpytest("-v", *plugin_args)
    # No warnings because patch should have been successful
    result.assert_outcomes(failed=1, passed=0, warnings=0)

    result.stdout.re_match_lines(
        (
            r"Expected: 'list([\n  1,\n  2,\n])'",
            # Actual is the original list's repr. No newlines or amber-style list prefix
            r"Actual: '[1, 2, 3]'",
        )
    )


@pytest.mark.filterwarnings("default")
def test_it_has_no_impact_on_non_syrupy_assertions(
    testdir, mock_teamcity_diff_tools, plugin_args
):
    # Generate diff and mimic EqualsAssertionError being thrown
    testdir.makepyfile(
        test_file="""

    def test_case():
        try:
            assert [1, 3] == [1, 2, 3]
        except:
            from teamcity.diff_tools import EqualsAssertionError

            err = EqualsAssertionError(expected=[1,3], actual=[1,2,3])
            print("Expected:", repr(str(err.expected)))
            print("Actual:", repr(str(err.actual)))
            raise
    """
    )

    result = testdir.runpytest("-v", *plugin_args)
    # No warnings because patch should have been successful
    result.assert_outcomes(failed=1, passed=0, warnings=0)

    result.stdout.re_match_lines(
        (
            r"Expected: '[1, 3]'",
            r"Actual: '[1, 2, 3]'",
        )
    )


@pytest.mark.filterwarnings("default")
def test_has_no_impact_on_real_exceptions_that_are_not_assertion_errors(
    testdir, mock_teamcity_diff_tools, plugin_args
):
    # Generate diff and mimic EqualsAssertionError being thrown
    testdir.makepyfile(
        test_file="""

    def test_case():
        try:
            assert [1, 3] == (1/0)
        except Exception as err:
            from teamcity.diff_tools import EqualsAssertionError

            err = EqualsAssertionError(
                expected=[1,3],
                actual=[1,2,3],
                real_exception=err
            )
            print("Expected:", repr(str(err.expected)))
            print("Actual:", repr(str(err.actual)))
            raise
    """
    )

    result = testdir.runpytest("-v", *plugin_args)
    # No warnings because patch should have been successful
    result.assert_outcomes(failed=1, passed=0, warnings=0)

    result.stdout.re_match_lines(
        (
            r"Expected: '[1, 3]'",
            r"Actual: '[1, 2, 3]'",
        )
    )



import pytest


@pytest.fixture
def testfile(testdir) -> pytest.Testdir:
    testdir.makepyfile(
        test_file=(
            """
            def test_case(snapshot):
                assert snapshot == "some-value"
            """
        ),
    )
    return testdir


def test_diff_mode_disabled_does_not_print_diff(
    testfile,
):
    # Generate initial snapshot
    result = testfile.runpytest("-v", "--snapshot-update")
    result.stdout.re_match_lines((r"1 snapshot generated\.",))
    assert result.ret == 0

    # Modify snapshot to generate diff
    testfile.makepyfile(
        test_file=(
            """
            def test_case(snapshot):
                assert snapshot == "some-other-value"
            """
        ),
    )

    # With diff we expect to see "some-other-value"
    result = testfile.runpytest("-v", "--snapshot-diff-mode=detailed")
    result.stdout.re_match_lines(
        (
            r".*- 'some-value'",
            r".*\+ 'some-other-value'",
        )
    )
    assert result.ret == 1

    # Without diff we do not expect to see "some-other-value"
    result = testfile.runpytest("-v", "--snapshot-diff-mode=disabled")
    result.stdout.no_re_match_line(r".*- 'some-value'")
    result.stdout.no_re_match_line(r".*\+ 'some-other-value'")
    assert result.ret == 1



"""Dataclass plugin for Syrupy Amber serializer."""

import dataclasses
from typing import Any

from syrupy.extensions.amber.serializer import (
    AmberDataSerializer,
    AmberDataSerializerPlugin,
    attr_getter,
)
from syrupy.types import SerializableData


class DataclassPlugin(AmberDataSerializerPlugin):
    """A Syrupy extension that serializes dataclass instances using Amber format."""

    @classmethod
    def is_data_serializable(cls, data: "SerializableData") -> bool:
        """Check if the data is a dataclass instance.

        Note: Excludes dataclass types themselves, which also yield True for is_dataclass.
        """
        return dataclasses.is_dataclass(data) and not isinstance(data, type)

    @classmethod
    def serialize(cls, data: "SerializableData", **kwargs: Any) -> str:
        """Serialize a dataclass instance into Amber format."""
        keys = sorted([f.name for f in dataclasses.fields(data)])
        return AmberDataSerializer.serialize_custom_iterable(
            data=data,
            resolve_entries=(keys, attr_getter, None),
            separator="=",
            **kwargs,
        )



def test_ignores_non_function_nodes(testdir, plugin_args):
    conftest = """
        import pytest

        class CustomItem(pytest.Item):
            def __init__(self, name, **kwargs):
                super().__init__(name, **kwargs)
                self._nodeid += "::CUSTOM"

            def runtest(self):
                pass

        def pytest_collect_file(file_path, parent):
            return CustomItem.from_parent(
                name=file_path.name,
                path=file_path,
                parent=parent
            )
        """
    testcase = """
        def test_example(snapshot):
            assert snapshot == 1
        """
    testdir.makeconftest(conftest)
    testdir.makepyfile(test_file=testcase)
    result = testdir.runpytest("test_file.py", "-v", "--snapshot-update", *plugin_args)
    result.stdout.re_match_lines((r".*test_file.py::CUSTOM.*",))
    assert result.ret == 0


def test_handles_pyargs_non_module_when_both_given(testdir, plugin_args):
    testdir.makeconftest("")
    testcase = """
        def test_example(snapshot):
            assert snapshot == 1
        """
    testdir.makepyfile(test_file=testcase)
    result = testdir.runpytest(
        "-v",
        "test_file.py",
        "--pyargs",
        "test_file",
        "--snapshot-update",
        *plugin_args,
    )
    assert result.ret == 0


def test_does_not_print_empty_snapshot_report(testdir, plugin_args_fails_xdist):
    testdir.makeconftest("")
    testcase_no_snapshots = """
        def test_example(snapshot):
            assert 1
        """
    testcase_yes_snapshots = """
        def test_example(snapshot):
            assert snapshot == 1
        """
    testdir.makepyfile(
        test_file_no=testcase_no_snapshots, test_file_yes=testcase_yes_snapshots
    )

    result = testdir.runpytest(
        "-v", "test_file_no.py", "--snapshot-update", *plugin_args_fails_xdist
    )
    result.stdout.re_match_lines((r".*test_file_no.py.*",))
    assert "snapshot report" not in result.stdout.str()
    assert "test_file_yes" not in result.stdout.str()
    assert result.ret == 0

    result = testdir.runpytest(
        "-v", "test_file_yes.py", "--snapshot-update", *plugin_args_fails_xdist
    )
    result.stdout.re_match_lines((r".*test_file_yes.py.*", r".*snapshot report.*"))
    assert result.ret == 0



import os
from datetime import datetime

from invoke import task

from .utils import ctx_run


@task
def install(ctx, upgrade=False):
    """
    Install dependencies and update lock file.
    """
    if upgrade:
        ctx_run(ctx, "poetry update")
    else:
        ctx_run(ctx, "poetry lock")
        ctx_run(ctx, "poetry install")


@task
def clean(ctx):
    """
    Remove build files e.g. package, distributable, compiled etc.
    """
    ctx_run(ctx, "rm -rf *.egg-info dist build __pycache__ .pytest_cache artifacts/*")


def version_scheme(v):
    if v.exact:
        return v.format_with("{tag}")
    return datetime.now().strftime("%Y.%m.%d.%H%M%S%f")


@task(pre=[clean])
def dist(ctx):
    """
    Generate version from scm and build package distributable
    """
    from setuptools_scm import get_version

    version = get_version(version_scheme=version_scheme, local_scheme=lambda _: "")
    ctx_run(ctx, f"poetry version {version}")
    ctx_run(ctx, "poetry build")


@task
def publish(ctx, dry_run=True):
    """
    Upload built package to pypi
    """
    repo_url = "--repository-url https://test.pypi.org/legacy/" if dry_run else ""
    ctx_run(ctx, f"twine upload --skip-existing {repo_url} dist/*")


@task(pre=[dist])
def release(ctx, dry_run=True, version=None):
    """
    Build and publish package to pypi index based on scm version
    """
    from semver.version import Version

    if not dry_run and not os.environ.get("CI"):
        print("This is a CI only command")
        exit(1)

    if dry_run and not version:
        version = ctx_run(ctx, "poetry version --short").stdout.strip()

    if not version:
        print("Missing version.")
        exit(1)

    try:
        Version.parse(version)
        should_publish_to_pypi = not dry_run
    except ValueError:
        should_publish_to_pypi = False

    # publish to test to verify builds
    if dry_run:
        publish(ctx, dry_run=True)

    # publish to pypi if test succeeds
    if should_publish_to_pypi:
        publish(ctx, dry_run=False)



from syrupy.extensions.amber.serializer import AmberDataSerializer
from syrupy.filters import props
from syrupy.matchers import path_type


class MyCustomClass:
    prop1 = 1
    prop2 = "a"
    prop3 = {1, 2, 3}


def test_snapshot_custom_class(snapshot):
    assert MyCustomClass() == snapshot


class MyCustomReprClass(MyCustomClass):
    def __repr__(self):
        state = "\n".join(
            f"  {a}={repr(getattr(self, a))},"
            for a in sorted(dir(self))
            if not a.startswith("_")
        )
        return f"{self.__class__.__name__}(\n{state}\n)"


def test_snapshot_custom_repr_class(snapshot):
    assert MyCustomReprClass() == snapshot


def test_snapshot_object_as_named_tuple_class(snapshot):
    """
    Show helper `object_as_named_tuple` to revert representation to amber standard
    """
    assert MyCustomReprClass() == snapshot(
        exclude=props("prop1"),
        matcher=path_type(
            types=(MyCustomReprClass,),
            replacer=lambda data, _: AmberDataSerializer.object_as_named_tuple(data),
        ),
    )



def test_no_failure_printed_if_all_failures_xfailed(testdir, plugin_args):
    testdir.makepyfile(
        test_file=(
            """
        import pytest

        @pytest.mark.xfail(reason="Failure expected.")
        def test_a(snapshot):
            assert snapshot == 'does-not-exist'
        """
        )
    )
    result = testdir.runpytest("-v", *plugin_args)
    result.stdout.no_re_match_line(r".*snapshot failed*")
    assert result.ret == 0


def test_failures_printed_if_only_some_failures_xfailed(
    testdir, plugin_args_fails_xdist
):
    testdir.makepyfile(
        test_file=(
            """
        import pytest

        @pytest.mark.xfail(reason="Failure expected.")
        def test_a(snapshot):
            assert snapshot == 'does-not-exist'

        def test_b(snapshot):
            assert snapshot == 'other'
        """
        )
    )
    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r".*1 snapshot failed*",))
    result.stdout.re_match_lines((r".*1 snapshot xfailed*",))
    assert result.ret == 1


def test_failure_printed_if_xfail_does_not_run(testdir, plugin_args_fails_xdist):
    testdir.makepyfile(
        test_file=(
            """
        import pytest

        @pytest.mark.xfail(False, reason="Failure expected.")
        def test_a(snapshot):
            assert snapshot == 'does-not-exist'
        """
        )
    )
    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r".*1 snapshot failed*",))
    result.stdout.no_re_match_line(r".*1 snapshot xfailed*")
    assert result.ret == 1



# noqa: A005
from collections.abc import Callable, Hashable
from typing import (
    Any,
)

SnapshotIndex = int | str
SerializableData = Any
SerializedData = str | bytes
PropertyName = Hashable
PropertyValueType = type[SerializableData]
PropertyPathEntry = tuple[PropertyName, PropertyValueType]
PropertyPath = tuple[PropertyPathEntry, ...]
try:
    from mypy_extensions import NamedArg

    PropertyMatcher = Callable[
        [
            NamedArg(SerializableData, "data"),  # noqa: F821
            NamedArg(PropertyPath, "path"),  # noqa: F821
        ],
        SerializableData | None,
    ]
    PropertyFilter = Callable[
        [
            NamedArg(PropertyName, "prop"),  # noqa: F821
            NamedArg(PropertyPath, "path"),  # noqa: F821
        ],
        bool,
    ]

except ImportError:
    globals()["PropertyMatcher"] = Callable[..., SerializableData | None]
    globals()["PropertyFilter"] = Callable[..., bool]



from pathlib import Path

import pytest


@pytest.fixture
def testfile(testdir):
    testdir.makepyfile(
        test_file=(
            """
            def test_default(snapshot):
                assert b"default extension serializer" == snapshot
            """
        ),
    )
    return testdir


def test_snapshot_default_extension_option_success(testfile):
    result = testfile.runpytest(
        "-v",
        "--snapshot-update",
        "--snapshot-default-extension",
        "syrupy.extensions.single_file.SingleFileSnapshotExtension",
    )
    result.stdout.re_match_lines((r"1 snapshot generated\.",))
    assert Path(
        testfile.tmpdir, "__snapshots__", "test_file", "test_default.raw"
    ).exists()
    assert not result.ret


def test_snapshot_default_extension_option_failure(testfile):
    result = testfile.runpytest(
        "-v",
        "--snapshot-update",
        "--snapshot-default-extension",
        "syrupy.extensions.amber.DoesNotExistExtension",
    )
    result.stdout.re_match_lines((r".*: Member 'DoesNotExistExtension' not found.*",))
    assert not Path(
        testfile.tmpdir, "__snapshots__", "test_file", "test_default.raw"
    ).exists()
    assert result.ret



import textwrap
from pathlib import Path

import pytest

import syrupy

SUBDIR = "subdir_not_on_default_path"


@pytest.fixture(autouse=True)
def cache_clear():
    syrupy.__import_extension.cache_clear()


@pytest.fixture
def testfile(pytester):
    subdir = pytester.mkpydir(SUBDIR)

    Path(
        subdir,
        "extension_file.py",
    ).write_text(
        data=textwrap.dedent(
            """
            from syrupy.extensions.single_file import SingleFileSnapshotExtension
            class MySingleFileExtension(SingleFileSnapshotExtension):
                pass
            """
        ),
        encoding="utf-8",
    )

    pytester.makepyfile(
        test_file=(
            """
            def test_default(snapshot):
                assert b"default extension serializer" == snapshot
            """
        )
    )

    return pytester


def test_snapshot_default_extension_option_success(testfile):
    testfile.makeini(
        f"""
        [pytest]
        pythonpath =
            {Path(testfile.path, SUBDIR).as_posix()}
    """
    )

    result = testfile.runpytest(
        "-v",
        "--snapshot-update",
        "--snapshot-default-extension",
        "extension_file.MySingleFileExtension",
    )
    result.stdout.re_match_lines((r"1 snapshot generated\.",))
    assert Path(
        testfile.path, "__snapshots__", "test_file", "test_default.raw"
    ).exists()
    assert not result.ret


def test_snapshot_default_extension_option_module_not_found(testfile):
    result = testfile.runpytest(
        "-v",
        "--snapshot-update",
        "--snapshot-default-extension",
        "extension_file.MySingleFileExtension",
    )
    result.stdout.re_match_lines((r".*: Module 'extension_file' does not exist.*",))
    assert not Path(
        testfile.path, "__snapshots__", "test_file", "test_default.raw"
    ).exists()
    assert result.ret


def test_snapshot_default_extension_option_failure(testfile):
    testfile.makeini(
        f"""
        [pytest]
        pythonpath =
            {Path(testfile.path, SUBDIR).as_posix()}
    """
    )

    result = testfile.runpytest(
        "-v",
        "--snapshot-update",
        "--snapshot-default-extension",
        "extension_file.DoesNotExistExtension",
    )
    result.stdout.re_match_lines((r".*: Member 'DoesNotExistExtension' not found.*",))
    assert not Path(
        testfile.path, "__snapshots__", "test_file", "test_default.raw"
    ).exists()
    assert result.ret



# type: ignore


def test_standard(testdir, benchmark):
    test_contents = """
        import base64
        from collections import namedtuple

        import pytest

        from syrupy.extensions.image import (
            PNGImageSnapshotExtension,
            SVGImageSnapshotExtension,
        )

        example_cycle_list = [1, 2, 3]
        example_cycle_list.append(example_cycle_list)
        example_cycle_dict = {"a": 1, "b": 2, "c": 3}
        example_cycle_dict.update(d=example_cycle_dict)
        example_png = base64.b64decode(
            b"iVBORw0KGgoAAAANSUhEUgAAADIAAAAyBAMAAADsEZWCAAAAG1BMVEXMzMy"
            b"Wlpaqqqq3t7exsbGcnJy+vr6jo6PFxcUFpPI/AAAACXBIWXMAAA7EAAAOxA"
            b"GVKw4bAAAAQUlEQVQ4jWNgGAWjgP6ASdncAEaiAhaGiACmFhCJLsMaIiDAE"
            b"QEi0WXYEiMCOCJAJIY9KuYGTC0gknpuHwXDGwAA5fsIZw0iYWYAAAAASUVO"
            b"RK5CYII="
        )
        example_svg = (
            '<?xml version="1.0" encoding="UTF-8"?>'
            '<svg viewBox="0 0 50 50" xmlns="http://www.w3.org/2000/svg">'
            '<g><rect width="50" height="50" fill="#fff"/>'
            '<g><g fill="#fff" stroke="#707070">'
            '<rect width="50" height="50" stroke="none"/>'
            '<rect x="0" y="0" width="50" height="50" fill="none"/></g>'
            '<text transform="translate(10 27)" fill="#707070" '
            'font-family="ConsolasForPowerline, Consolas for Powerline" font-size="8">'
            '<tspan x="0" y="0">50 x 50</tspan></text></g></g></svg>'
        )
        ExampleTuple = namedtuple("ExampleTuple", ["a", "b", "c", "d"])


        class ExampleClass:
            a = 1
            b = "2"
            c = example_cycle_list
            d = example_cycle_dict
            _protected_variable = None
            __private_variable = None

            def __init__(self, x=None):
                self.x = x
                self._y = 1
                self.__z = 2

            def public_method(self, a, b=1, *, c, d=None):
                pass

            def _protected_method(self):
                pass

            def __private_method(self):
                pass


        test_cases = [
            False,
            True,
            3.5,
            7,
            2 / 6,
            22 / 7,
            "",
            r"Raw string",
            r"Escaped \\n",
            r"Backslash \\u U",
            "🥞🐍🍯",
            "singleline:",
            "- singleline",
            "line 1\\nline 2\\nline 3",
            "line 2\\nline 2\\n  line 3",
            "line 1\\r\\nline 2\\r\\nline 3",
            "string with 'quotes'",
            b"Byte string",
            ("this", "is", ("a", "tuple")),
            ExampleTuple(a="this", b="is", c="a", d={"named", "tuple"}),
            {"this", "is", "a", "set"},
            {"contains", "frozen", frozenset({"1", "2"})},
            {"contains", "tuple", (1, 2)},
            {"contains", "namedtuple", ExampleTuple(a=1, b=2, c=3, d=4)},
            {"b": True, "c": "Some text.", "d": ["1", 2], "a": {"e": False}},
            {"b": True, "c": "Some ttext.", "d": ["1", 2], "a": {"e": False}},
            {
                1: True,
                "a": "Some ttext.",
                frozenset({"1", "2"}): ["1", 2],
                ExampleTuple(a=1, b=2, c=3, d=4): {"e": False},
            },
            [1, 2, "string", {"key": "value"}],
            example_cycle_list,
            example_cycle_dict,
            ExampleClass(ExampleClass()),
        ]


        class TestClass:
            def test_method(self, snapshot):
                assert example_svg == snapshot(
                    extension_class=SVGImageSnapshotExtension
                )
                assert example_svg == snapshot
                assert example_png == snapshot(
                    extension_class=PNGImageSnapshotExtension
                )

            @pytest.mark.parametrize(
                "actual",
                test_cases,
                ids=lambda x: "",
            )
            def test_parametrized_method(self, snapshot, actual):
                assert actual == snapshot
                snapshot.assert_match(actual)
                assert snapshot == actual

            @pytest.mark.parametrize(
                "actual",
                test_cases,
                ids=lambda x: "",
            )
            class TestNestedParametrizedClass:
                def test_nested_method(self, snapshot, actual):
                    assert snapshot == actual


        @pytest.mark.parametrize("x", range(1000))
        def test_performance(x, snapshot):
            assert x == snapshot
        """
    testdir.makepyfile(test=test_contents)

    # Run benchmark
    benchmark(lambda: testdir.runpytest("test.py", "--snapshot-update"))



import sys
from pathlib import Path

import pytest


@pytest.fixture
def testcases_initial():
    return {
        "test_used": (
            """
            import pytest

            @pytest.mark.parametrize("actual", [1, 2, 3])
            def test_used(snapshot, actual):
                assert snapshot == actual

            def test_used1(snapshot):
                assert snapshot == 'unused'
                assert 'unused' == snapshot
            """
        ),
        "test_updated_1": (
            """
            def test_updated_1(snapshot):
                assert snapshot == ['this', 'will', 'be', 'updated']
            """
        ),
        "test_updated_2": (
            """
            def test_updated_2(snapshot):
                assert ['this', 'will', 'be', 'updated'] == snapshot
            """
        ),
        "test_updated_3": (
            """
            def test_updated_3(snapshot):
                assert snapshot == ['this', 'will', 'be', 'updated']
            """
        ),
        "test_updated_4": (
            """
            def test_updated_4(snapshot):
                assert snapshot == "single line change"
            """
        ),
        "test_updated_5": (
            """
            def test_updated_5(snapshot):
                assert snapshot == '''
                multiple line changes
                with some lines staying the same
                intermittent changes that have to be ignore by the differ output
                because when there are a lot of changes you only want to see changes
                you do not want to see this line
                or this line
                this line should not show up because new lines are normalised\\r\\n
                \x1b[38;5;1mthis line should show up because it changes color\x1b[0m
                '''
            """
        ),
    }


@pytest.fixture
def testcases_updated(testcases_initial):
    updates = {
        "test_updated_1": (
            """
            def test_updated_1(snapshot):
                assert snapshot == ['this', 'will', 'not', 'match']
            """
        ),
        "test_updated_2": (
            """
            def test_updated_2(snapshot):
                assert ['this', 'will', 'fail'] == snapshot
            """
        ),
        "test_updated_3": (
            """
            def test_updated_3(snapshot):
                assert snapshot == ['this', 'will', 'be', 'too', 'much']
            """
        ),
        "test_updated_4": (
            """
            def test_updated_4(snapshot):
                assert snapshot == "sing line changeling"
            """
        ),
        "test_updated_5": (
            """
            def test_updated_5(snapshot):
                assert snapshot == '''
                multiple line changes
                with some lines not staying the same
                intermittent changes so unchanged lines have to be ignored by the differ
                cause when there are a lot of changes you only want to see what changed
                you do not want to see this line
                or this line
                this line should not show up because new lines are normalised\\n
                \x1b[38;5;3mthis line should show up because it changes color\x1b[0m
                and this line does not exist in the first one
                '''
            """
        ),
    }
    return {**testcases_initial, **updates}


@pytest.fixture
def run_testcases(testdir, testcases_initial):
    sys.path.append(str(testdir.tmpdir))
    testdir.makepyfile(**testcases_initial)
    result = testdir.runpytest("-v", "--snapshot-update")
    result.stdout.re_match_lines((r"10 snapshots generated.",))
    assert "Can not relate snapshot name" not in result.stdout.str()

    return result, testdir, testcases_initial


def test_update_failure_shows_snapshot_diff(
    run_testcases, testcases_updated, plugin_args_fails_xdist
):
    testdir = run_testcases[1]
    testdir.makepyfile(**testcases_updated)
    result = testdir.runpytest("-vv", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(
        (
            r".*assert snapshot == \['this', 'will', 'not', 'match'\]",
            r".*AssertionError: assert \[- snapshot\] == \[\+ received\]",
            r".*    list\(\[",
            r".*     ...",
            r".*      'will',",
            r".*  -   'be',",
            r".*  -   'updated',",
            r".*  \+   'not',",
            r".*  \+   'match',",
            r".*    \]",
            r".*assert \['this', 'will', 'fail'\] == snapshot",
            r".*AssertionError: assert \[\+ received\] == \[- snapshot\]",
            r".*    list\(\[",
            r".*     ...",
            r".*      'will',",
            r".*  -   'be',",
            r".*  \+   'fail',",
            r".*  -   'updated',",
            r".*    \]",
            r".*assert snapshot == \['this', 'will', 'be', 'too', 'much'\]",
            r".*AssertionError: assert \[- snapshot\] == \[\+ received\]",
            r".*    list\(\[",
            r".*     ...",
            r".*      'be',",
            r".*  -   'updated',",
            r".*  \+   'too',",
            r".*  \+   'much',",
            r".*    \]",
            r".*assert snapshot == \"sing line changeling\"",
            r".*AssertionError: assert \[- snapshot\] == \[\+ received\]",
            r".*  - 'single line change'",
            r".*  \+ 'sing line changeling'",
            r".*AssertionError: assert \[- snapshot\] == \[\+ received\]",
            r".*    '",
            r".*      ...",
            r".*        multiple line changes",
            r".*  -     with some lines staying the same",
            r".*  \+     with some lines not staying the same",
            r".*  -     intermittent changes that have to be ignore by the differ out",
            r".*  \+     intermittent changes so unchanged lines have to be ignored b",
            r".*  -     because when there are a lot of changes you only want to see ",
            r".*  \+     cause when there are a lot of changes you only want to see w",
            r".*        you do not want to see this line",
            r".*      ...",
            r".*    ",
            r".*  -     \[38;5;1mthis line should show up because it changes color",
            r".*  \+     \[38;5;3mthis line should show up because it changes color",
            r".*  \+     and this line does not exist in the first one",
            r".*        ",
            r".*    '",
        )
    )
    assert result.ret == 1


def test_update_success_shows_snapshot_report(
    run_testcases, testcases_updated, plugin_args_fails_xdist
):
    testdir = run_testcases[1]
    testdir.makepyfile(**testcases_updated)
    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r"5 snapshots passed\. 5 snapshots updated\.",))
    assert result.ret == 0


def test_update_targets_only_selected_parametrized_tests_for_update_dash_m(
    run_testcases, plugin_args_fails_xdist
):
    updated_tests = {
        "test_used": (
            """
            import pytest

            @pytest.mark.parametrize("actual", [1, "2"])
            def test_used(snapshot, actual):
                assert snapshot == actual
            """
        ),
    }
    testdir = run_testcases[1]
    testdir.makepyfile(**updated_tests)
    result = testdir.runpytest(
        "-v", "--snapshot-update", *plugin_args_fails_xdist, "-m", "parametrize"
    )
    result.stdout.re_match_lines(
        (
            r"1 snapshot passed\. 1 snapshot updated\. 1 unused snapshot deleted\.",
            r"Deleted test_used\[3\] \(__snapshots__[\\/]test_used.ambr\)",
        )
    )
    snapshot_path = [testdir.tmpdir, "__snapshots__"]
    assert Path(*snapshot_path, "test_used.ambr").exists()
    assert Path(*snapshot_path, "test_updated_1.ambr").exists()


def test_update_targets_only_selected_parametrized_tests_for_update_dash_k(
    run_testcases, plugin_args_fails_xdist
):
    updated_tests = {
        "test_used": (
            """
            import pytest

            @pytest.mark.parametrize("actual", [1, "2", 3])
            def test_used(snapshot, actual):
                assert snapshot == actual
            """
        ),
    }
    testdir = run_testcases[1]
    testdir.makepyfile(**updated_tests)
    result = testdir.runpytest(
        "-v", "--snapshot-update", *plugin_args_fails_xdist, "-k", "test_used[2]"
    )
    result.stdout.re_match_lines((r"1 snapshot updated\.",))
    assert "Deleted" not in result.stdout.str()
    snapshot_path = [testdir.tmpdir, "__snapshots__"]
    assert Path(*snapshot_path, "test_used.ambr").exists()
    assert Path(*snapshot_path, "test_updated_1.ambr").exists()


def test_update_targets_only_selected_parametrized_tests_for_removal_dash_k(
    run_testcases, plugin_args_fails_xdist
):
    updated_tests = {
        "test_used": (
            """
            import pytest

            @pytest.mark.parametrize("actual", [1, 2])
            def test_used(snapshot, actual):
                assert snapshot == actual
            """
        ),
    }
    testdir = run_testcases[1]
    testdir.makepyfile(**updated_tests)
    result = testdir.runpytest(
        "-v", "--snapshot-update", *plugin_args_fails_xdist, "-k", "test_used["
    )
    result.stdout.re_match_lines(
        (
            r"2 snapshots passed\. 1 unused snapshot deleted\.",
            r"Deleted test_used\[3\] \(__snapshots__[\\/]test_used\.ambr\)",
        )
    )
    snapshot_path = [testdir.tmpdir, "__snapshots__"]
    assert Path(*snapshot_path, "test_used.ambr").exists()
    assert Path(*snapshot_path, "test_updated_1.ambr").exists()


def test_update_targets_only_selected_class_tests_dash_k(
    testdir, plugin_args_fails_xdist
):
    test_content = """
        import pytest

        class TestClass:
            def test_case_1(self, snapshot):
                assert snapshot == 1

            def test_case_2(self, snapshot):
                assert snapshot == 2
        """

    testdir.makepyfile(test_content=test_content)
    testdir.runpytest("-v", "--snapshot-update")
    assert Path(testdir.tmpdir, "__snapshots__", "test_content.ambr").exists()

    result = testdir.runpytest(
        "test_content.py", "-v", *plugin_args_fails_xdist, "-k", "test_case_2"
    )
    result.stdout.re_match_lines((r"1 snapshot passed\.",))
    assert "snapshot unused" not in result.stdout.str()


def test_update_targets_only_selected_module_tests_dash_k(
    testdir, plugin_args_fails_xdist
):
    test_content = """
        import pytest

        def test_case_1(snapshot):
            assert snapshot == 1

        def test_case_2(snapshot):
            assert snapshot == 2
        """

    testdir.makepyfile(test_content=test_content)
    testdir.runpytest("-v", "--snapshot-update")
    assert Path(testdir.tmpdir, "__snapshots__", "test_content.ambr").exists()

    result = testdir.runpytest(
        "test_content.py", "-v", *plugin_args_fails_xdist, "-k", "test_case_2"
    )
    result.stdout.re_match_lines((r"1 snapshot passed\.",))
    assert "snapshot unused" not in result.stdout.str()


def test_update_targets_only_selected_module_tests_nodes(
    run_testcases, plugin_args_fails_xdist
):
    testdir = run_testcases[1]
    snapfile_empty = Path("__snapshots__", "empty_snapfile.ambr")
    testdir.makefile(".ambr", **{str(snapfile_empty): ""})
    testfile = Path(testdir.tmpdir, "test_used.py")
    result = testdir.runpytest(
        "-v", f"{testfile}::test_used", "--snapshot-update", *plugin_args_fails_xdist
    )
    result.stdout.re_match_lines((r"3 snapshots passed\.",))
    assert "unused" not in result.stdout.str()
    assert "updated" not in result.stdout.str()
    assert "deleted" not in result.stdout.str()
    assert result.ret == 0
    assert snapfile_empty.exists()


def test_update_targets_only_selected_module_tests_nodes_pyargs(
    run_testcases, plugin_args_fails_xdist
):
    testdir = run_testcases[1]
    snapfile_empty = Path("__snapshots__", "empty_snapfile.ambr")
    testdir.makefile(".ambr", **{str(snapfile_empty): ""})
    result = testdir.runpytest(
        "-v",
        "--snapshot-update",
        *plugin_args_fails_xdist,
        "--pyargs",
        "test_used::test_used",
    )
    result.stdout.re_match_lines((r"3 snapshots passed\.",))
    assert "unused" not in result.stdout.str()
    assert "updated" not in result.stdout.str()
    assert "deleted" not in result.stdout.str()
    assert result.ret == 0
    assert snapfile_empty.exists()


def test_update_targets_only_selected_module_tests_file_for_update(
    run_testcases, plugin_args_fails_xdist
):
    testdir = run_testcases[1]
    snapfile_empty = Path("__snapshots__", "empty_snapfile.ambr")
    testdir.makefile(".ambr", **{str(snapfile_empty): ""})
    testdir.makepyfile(
        test_used=(
            """
            import pytest

            @pytest.mark.parametrize("actual", [1, 2, 3])
            def test_used(snapshot, actual):
                assert snapshot == actual
            """
        )
    )
    result = testdir.runpytest(
        "-v", "test_used.py", "--snapshot-update", *plugin_args_fails_xdist
    )
    result.stdout.re_match_lines(
        (
            r"3 snapshots passed\. 2 unused snapshots deleted\.",
            r"Deleted test_used1, test_used1\.1 \(__snapshots__[\\/]test_used\.ambr\)",
        )
    )
    assert result.ret == 0
    assert snapfile_empty.exists()
    assert Path("__snapshots__", "test_used.ambr").exists()


def test_update_targets_only_selected_module_tests_file_for_removal(
    run_testcases, plugin_args_fails_xdist
):
    testdir = run_testcases[1]
    testdir.makepyfile(
        test_used=(
            """
            def test_used(snapshot):
                assert True
            """
        ),
    )
    snapfile_empty = Path("__snapshots__", "empty_snapfile.ambr")
    testdir.makefile(".ambr", **{str(snapfile_empty): ""})
    result = testdir.runpytest(
        "-v", "test_used.py", "--snapshot-update", *plugin_args_fails_xdist
    )
    result.stdout.re_match_lines(
        (
            r"5 unused snapshots deleted\.",
            r"Deleted test_used1, test_used1\.1, test_used\[1\], test_used\[2\]"
            r", test_used\[3\] \(__snapshots__[\\/]test_used\.ambr\)",
        )
    )
    assert result.ret == 0
    assert snapfile_empty.exists()
    assert not Path("__snapshots__", "test_used.ambr").exists()


def test_update_removes_empty_snapshot_collection_only(
    run_testcases, plugin_args_fails_xdist
):
    testdir = run_testcases[1]
    snapfile_empty = Path("__snapshots__", "empty_snapfile.ambr")
    testdir.makefile(".ambr", **{str(snapfile_empty): ""})
    assert snapfile_empty.exists()
    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(
        (
            r"10 snapshots passed\. 1 unused snapshot deleted\.",
            r"Deleted empty snapshot collection "
            r"\(__snapshots__[\\/]empty_snapfile\.ambr\)",
        )
    )
    assert result.ret == 0
    assert not snapfile_empty.exists()
    assert Path("__snapshots__", "test_used.ambr").exists()


def test_update_removes_hanging_snapshot_collection_file(
    run_testcases, plugin_args_fails_xdist
):
    testdir = run_testcases[1]
    snapfile_used = Path("__snapshots__", "test_used.ambr")
    snapfile_hanging = Path("__snapshots__", "hanging_snapfile.abc")
    testdir.makefile(".abc", **{str(snapfile_hanging): ""})
    assert snapfile_hanging.exists()
    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(
        (
            r"10 snapshots passed\. 1 unused snapshot deleted\.",
            r"Deleted unknown snapshot collection "
            r"\(__snapshots__[\\/]hanging_snapfile\.abc\)",
        )
    )
    assert f"{snapfile_used}" not in result.stdout.str()
    assert result.ret == 0
    assert snapfile_used.exists()
    assert not snapfile_hanging.exists()



import pytest

from syrupy.extensions.amber import (
    AmberDataSerializerSorted,
    AmberSnapshotExtension,
)


class AmberSortedSnapshotExtension(AmberSnapshotExtension):
    serializer_class = AmberDataSerializerSorted


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(AmberSortedSnapshotExtension)


def test_many_sorted(snapshot):
    for i in range(25):
        assert i == snapshot



"""
Example: Custom Snapshot Name
"""

import pytest

from syrupy.extensions.amber import AmberSnapshotExtension
from syrupy.location import PyTestLocation
from syrupy.types import SnapshotIndex


class CanadianNameExtension(AmberSnapshotExtension):
    @classmethod
    def get_snapshot_name(
        cls, *, test_location: "PyTestLocation", index: "SnapshotIndex"
    ) -> str:
        original_name = AmberSnapshotExtension.get_snapshot_name(
            test_location=test_location, index=index
        )
        return f"{original_name}🇨🇦"


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(CanadianNameExtension)


def test_canadian_name(snapshot):
    assert "Name should be test_canadian_name🇨🇦." == snapshot



import pytest


@pytest.fixture
def testcases_initial(testdir):
    testdir.makeconftest(
        """
        import pytest
        import math

        from syrupy.extensions.amber import AmberSnapshotExtension

        class CustomSnapshotExtension(AmberSnapshotExtension):
            def matches(self, *, serialized_data, snapshot_data):
                try:
                    a = float(serialized_data)
                    b = float(snapshot_data)
                    return math.isclose(a, b, rel_tol=1e-5)
                except:
                    return False

        @pytest.fixture
        def snapshot_custom(snapshot):
            return snapshot.use_extension(CustomSnapshotExtension)
        """
    )
    return {
        "passed": (
            """
            def test_passed_custom(snapshot_custom):
                assert snapshot_custom == 3.0
            """
        ),
        "failed": (
            """
            def test_passed_custom(snapshot_custom):
                # this comment is required or the test breaks
                assert snapshot_custom == 4.0
            """
        ),
    }


@pytest.fixture
def generate_snapshots(testdir, testcases_initial):
    testdir.makepyfile(test_file=testcases_initial["passed"])
    result = testdir.runpytest("-v", "--snapshot-update")
    return result, testdir, testcases_initial


@pytest.mark.xfail(strict=False)
def test_generated_snapshots(generate_snapshots):
    result = generate_snapshots[0]
    result.stdout.re_match_lines((r"1 snapshot generated\.",))
    assert "snapshots unused" not in result.stdout.str()
    assert result.ret == 0


@pytest.mark.xfail(strict=False)
def test_approximate_match(generate_snapshots, plugin_args_fails_xdist):
    testdir = generate_snapshots[1]
    testdir.makepyfile(
        test_file="""
            def test_passed_custom(snapshot_custom):
                assert snapshot_custom == 3.2
            """
    )
    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r"test_file.py::test_passed_custom PASSED",))
    assert result.ret == 0


@pytest.mark.xfail(strict=False)
def test_failed_snapshots(generate_snapshots):
    testdir = generate_snapshots[1]
    testdir.makepyfile(test_file=generate_snapshots[2]["failed"])
    result = testdir.runpytest("-v")
    result.stdout.re_match_lines((r"1 snapshot failed\.",))
    assert result.ret == 1


@pytest.mark.xfail(strict=False)
def test_updated_snapshots(generate_snapshots, plugin_args_fails_xdist):
    _, testdir, initial = generate_snapshots
    testdir.makepyfile(test_file=initial["failed"])
    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r"1 snapshot updated\.",))
    assert result.ret == 0



""" "Tests for the Amber serializer plugins for attrs, dataclasses, and Pydantic models."""

from dataclasses import dataclass

import attr
import pytest
from pydantic import BaseModel

from syrupy.extensions.amber import AmberSnapshotExtension
from syrupy.extensions.amber.attrs_plugin import AttrsPlugin
from syrupy.extensions.amber.dataclasses_plugin import DataclassPlugin
from syrupy.extensions.amber.pydantic_plugin import PydanticPlugin
from syrupy.extensions.amber.serializer import AmberDataSerializer


# region Data Classes
@attr.s
class AttrsPoint:
    x = attr.ib()
    y = attr.ib()


@dataclass
class DataclassPoint:
    x: int
    y: int


class PydanticPoint(BaseModel):
    x: int
    y: int


class Inner(BaseModel):
    val: int


class ComplexModel(BaseModel):
    """Complex Pydantic model for testing nested structures.

    Note: make sure field names are not sorted alphabetically to test ordering.
    """

    values: list[Inner]

    start: PydanticPoint
    end: PydanticPoint

    mapping: dict[str, Inner]


@dataclass
class Middle:
    inner: Inner
    other: str


@attr.s
class Outer:
    middle = attr.ib()
    tags = attr.ib(factory=list)


# endregion


# region Serializers
class AttrsSerializer(AmberDataSerializer):
    serializer_plugins = [AttrsPlugin]


class DataclassSerializer(AmberDataSerializer):
    serializer_plugins = [DataclassPlugin]


class PydanticSerializer(AmberDataSerializer):
    serializer_plugins = [PydanticPlugin]


class MixedSerializer(AmberDataSerializer):
    serializer_plugins = [AttrsPlugin, DataclassPlugin, PydanticPlugin]


# endregion


# region Extensions
class AmberAttrsExtension(AmberSnapshotExtension):
    serializer_class = AttrsSerializer


class AmberDataclassExtension(AmberSnapshotExtension):
    serializer_class = DataclassSerializer


class AmberPydanticExtension(AmberSnapshotExtension):
    serializer_class = PydanticSerializer


class AmberMixedExtension(AmberSnapshotExtension):
    serializer_class = MixedSerializer


# endregion


# region Fixtures


@pytest.fixture
def snapshot_attrs(snapshot) -> AmberSnapshotExtension:
    return snapshot.use_extension(AmberAttrsExtension)


@pytest.fixture
def snapshot_dataclass(snapshot) -> AmberSnapshotExtension:
    return snapshot.use_extension(AmberDataclassExtension)


@pytest.fixture
def snapshot_pydantic(snapshot) -> AmberSnapshotExtension:
    return snapshot.use_extension(AmberPydanticExtension)


@pytest.fixture
def snapshot_mixed(snapshot) -> AmberSnapshotExtension:
    return snapshot.use_extension(AmberMixedExtension)


# endregion


# region Tests
def test_attrs_plugin(snapshot_attrs):
    """Test serialization of an attrs class using the AmberAttrsExtension."""
    point = AttrsPoint(x=1, y=2)
    assert point == snapshot_attrs


def test_dataclasses_plugin(snapshot_dataclass):
    """Test serialization of a dataclass using the AmberDataclassExtension."""
    point = DataclassPoint(x=1, y=2)
    assert point == snapshot_dataclass


def test_pydantic_plugin(snapshot_pydantic):
    """Test serialization of a Pydantic model using the AmberPydanticExtension."""
    point = PydanticPoint(x=1, y=2)
    assert point == snapshot_pydantic


def test_mixed_plugins(snapshot_mixed):
    """Test serialization of mixed data types using the AmberMixedExtension."""
    complex_data = {
        "attrs": AttrsPoint(x=1, y=2),
        "dataclass": DataclassPoint(x=3, y=4),
        "pydantic": PydanticPoint(x=5, y=6),
        "list_mixed": [
            AttrsPoint(x=10, y=20),
            DataclassPoint(x=30, y=40),
        ],
    }
    assert complex_data == snapshot_mixed


def test_nested_structures(snapshot_mixed):
    """Test serialization of nested structures using the AmberMixedExtension."""
    data = Outer(middle=Middle(inner=Inner(val=42), other="nested"), tags=["c"])
    assert data == snapshot_mixed


def test_complex_pydantic_model(snapshot_pydantic):
    """Test serialization of a complex Pydantic model using the AmberPydanticExtension.

    Ensures nested models, lists, and dicts are handled correctly.
    Ensures field ordering is preserved as defined in the model.
    """
    data = ComplexModel(
        end=PydanticPoint(x=10, y=10),
        mapping={"first": Inner(val=100), "second": Inner(val=200)},
        start=PydanticPoint(x=0, y=0),
        values=[Inner(val=1), Inner(val=2)],
    )
    assert data == snapshot_pydantic


# endregion



from pathlib import Path


def test_ignore_file_extensions(testdir):
    # Generate initial snapshot file
    testdir.makepyfile(
        test_file=(
            """
            def test_case(snapshot):
                assert snapshot == "some-value"
            """
        ),
    )
    result = testdir.runpytest("-v", "--snapshot-update")
    result.stdout.re_match_lines((r"1 snapshot generated\.",))
    assert result.ret == 0

    # Duplicate the snapshot file with "dvc" file extension
    snapshot_file = Path(testdir.tmpdir, "__snapshots__", "test_file.ambr")
    dvc_file = snapshot_file.with_suffix(".ambr.dvc")
    dvc_file.write_text(snapshot_file.read_text())

    # Run with ignored file extension
    result = testdir.runpytest(
        "-v", "--snapshot-details", "--snapshot-ignore-file-extensions=dvc"
    )
    result.stdout.no_re_match_line(r".*1 snapshot unused.*")
